initSidebarItems({"trait":[["DeltaEq","Trait to determine whether two values are within a certain tolerance of [`DeltaE`]. Types that implement Into<[`LabValue`]> implicitly implement [`Delta`]. Types that implement [`Delta`] and [`Copy`] may also implement DeltaEq for other types that also implement [`Delta`] and [`Copy`]."],["Tolerance","Trait to define a tolerance value for the [`DeltaEq`] trait"]]});